/*
    Trabalho Pratico II - IPDI
    -------------------------
    Anderson de Rezende Rocha <anderson.rocha@ic.unicamp.br>
    Leyza E. Baldo Dorini <leyzad@yahoo.com.br>

    Deteccao de circulos em imagens atraves da transformada de HOUGH
    Baseado no algoritmo proposto por Duda e Hart
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "imagem.h"
#include "morfo.h"
#include "utils.h"
#include "tratadorErros.h"

// Valor de PI = Comprimento de um circulo dividido por seu raio
#define PI 3.141592654

// Rotina para achar circulos na imagem de entrada atraves da transformada de Hough
void detectarCirculos(int *imagem, int *imagemSaida, int nl, int nc);

int main (int argc, char *argv[])
{
    int OK, nc, nl, mn;
    int *imagemEntrada;
    int *imagemSaida;

    printf("\nDeteccao de circulos pela transformada de Hough <Anderson Rocha, Leyza Dorini>\n");
    if (argc == 3)
    {
        OK = le_imagem_pgm (argv[1], &imagemEntrada, &nl, &nc, &mn);

        info_imagem (argv[1], nl, nc, mn);
        aloca_memo (&imagemSaida, nl, nc);
        
        printf("Indo para a funo detectar circulos!! \n");

        detectarCirculos(imagemEntrada, imagemSaida, nl, nc);

        if(imagemEntrada != NULL)
        {
            grava_imagem_pgm (imagemSaida, argv[2], nl, nc, mn);
            printf("Imagem \"%s\" gravada com sucesso!\n", argv[2]);
        }

        desaloca_memo(&imagemSaida);
    }
    else
    {
        tratarSintaxe(7);
    }

    printf("\n");
    return(0);
}

// Rotina para achar circulos na imagem de entrada atraves da transformada de Hough
void detectarCirculos(int *imagemEntrada, int *imagemSaida, int nl, int nc)
{
    // Tamanho do plano de cada parametro
    // Escolhe o maior entre nl e nc para representar o EP
    int TAM_EP = (nl > nc ? nl : nc);
    // Espaco de Parametros tridimensional para acomodar as coordenadas
    // do centro do circulo (xc, yc) e o raio respectivo. No espaco sera
    // entao (xc, yc, r)    
    int EP[TAM_EP][TAM_EP][TAM_EP];
    // Matriz para acomodar a imagem lida. Achamos interessante trabalhar com a imagem
    // em forma matricial.
    int imagem[nl][nc];
    // Indice para as linhas e colunas e operacoes
    // x central, y central e raio de um circulo
    int i, j, k, xc, yc, raio;
    // Auxiliares no calculo do acumulador com maior valor agregado
    int limiar, xc_maior,  yc_maior;
    // Vator de auxilio para desenhar um circulo com suas coordenadas polares
    int vet_aux [2];
    // Auxiliar que indica o tamanho do passo no calculo das coordenadas polares
    float aux = 0.0;

    // Melhor trabalhar diretamente com matrizes. Logo, devemos passar o contedo de imagemEntrada que  um
    // ponteiro, para a matriz imagem[][]
    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
        {
            // Le cada pixel e passa para a matriz
            imagem[i][j] = *(imagemEntrada + i * nc + j);
        }
    }

    // Inicializando o Espaco de Parametros
    for(i = 0; i < TAM_EP; i++)
    {
        for(j = 0; j < TAM_EP; j++)
        {
            for(k = 0; k < TAM_EP; k++)
            {
                EP[i][j][k] = 0;
            }
        }
    }

    //preenchendo o espao de parmetros
    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
        {
         if(imagem[i][j] == 255)
            {
                // Dado um ponto x,y da imagem achar no espaco de parametros (EP)
                // qual o raio correspondente
                for(xc = 0; xc < TAM_EP; xc++)
                {
                    for(yc = 0; yc < TAM_EP; yc++)
                    {
                        raio = (int)sqrt(pow((i - xc), 2) + pow((j - yc),2));

                        if(raio < TAM_EP)
                        {
                           // Incrementar o acumulador pois o ponto i, j da imagem pode
                           // representar um circulo com centro em xc,yc e raio k
                           // no Espaco de Parametros
                           EP[xc][yc][raio] += 1;
                        }
                    }
                }
            }
        }
    }

    // Achando o acumulador com maior valor
    // Este valor  empirico
    limiar = 0;
    xc_maior = 0;
    yc_maior = 0;

    // As observacoes abaixo sao apenas empiricas. Nao ha fundamentacao teorica
    // R = 12 -> Limiar = 42
    // R = 18 -> Limiar = 72
    // R = 24 -> Limiar = 96

    // Pode ser resumido pela formula:
    // L = [piso(R / 2 + 1)] * 5

    for(i = 0; i < TAM_EP; i++)
    {
        for(j = 0; j < TAM_EP; j++)
        {
            for(k = 0; k < TAM_EP; k++) 
            {
                // Calculo empirico do limiar

                //  desnecessrio fazer calculos para raios muito pequenos pois
                // eles geralmente no representam nada em uma figura real. Por definicao
                // o nosso menor circulo detectavel tem raio 10
                limiar = (k / 2 + 1) * 5;

                if(EP[i][j][k] > limiar)
                {
                    xc_maior = i;
                    yc_maior = j;
                    raio = k;
                    aux = 0.0;

                    while(aux <= (2*PI))
                    {
                        vet_aux[0] = xc_maior  + raio*cos(aux);
                        vet_aux[1] = yc_maior  + raio*sin(aux);
                        aux = aux + 0.01;

                        if((vet_aux[0]>0)&&(vet_aux[1]>0))
                        {
                            if((vet_aux[0] * nc + vet_aux[1]) < (nl * nc))
                            {
                                *(imagemSaida + vet_aux[0] * nc + vet_aux[1]) = 255;
                            }
                        }
                    }
                }
            }
        }
    }    
}
